import sys
import os


def get_user_app_data():
    platform = sys.platform
    if platform.startswith('linux'):
        return os.getenv('HOME')
    elif platform.startswith('win32'):
        return os.getenv('LOCALAPPDATA')
    else:
        raise SystemError(f'The platform {platform} is not supported.')



import aiohttp
import logging
from vapordmods.api.base import BaseApi

api_logger = logging.getLogger(__name__)


class thunderstore(BaseApi):

    _THUNDERSTORE_API_URL_LATEST = 'https://thunderstore.io/api/experimental/package/{}/{}/'
    _THUNDERSTORE_API_URL_VERSION = 'https://thunderstore.io/api/experimental/package/{}/{}/{}'
    _THUNDERSTORE_DOWNLOAD_LINK = 'https://gcdn.thunderstore.io/live/repository/packages/{}'

    def __init__(self):
        super().__init__()

    async def get_update(self, namespace: str, name: str, mods_dir: str, version: str = None, api_key: str = None) -> int:
        if not version:
            request = self._THUNDERSTORE_API_URL_LATEST.format(namespace, name)
        else:
            request = self._THUNDERSTORE_API_URL_VERSION.format(namespace, name, version)

        async with aiohttp.request('GET', request) as resp:
            if resp.status == 200:
                j = await resp.json()
                if not version:
                    self.version = j['latest']['version_number']
                    self.description = j['latest']['description']
                    download_url = j['latest']['full_name'] + '.zip'
                else:
                    self.version = j['version_number']
                    self.description = j['description']
                    download_url = j['full_name'] + '.zip'

                self.provider = 'thunderstore'
                self.app = namespace
                self.mods = name
                self.title = name
                self.mods_dir = mods_dir
                self.full_mods_name = namespace + '-' + name

                self.download_url = self._THUNDERSTORE_DOWNLOAD_LINK.format(download_url)
                api_logger.debug(
                    f'The request from the "Thunderstore" API was successfull for the namespace {namespace} and the mod {name}')
                return 0
            else:
                api_logger.error(f'{namespace}-{name}: Status {resp.status}, Error: {await resp.text()}')
                return 1



import aiohttp
import logging

from vapordmods.api.base import BaseApi

api_logger = logging.getLogger(__name__)


class workshop(BaseApi):

    _WORKSHOP_API_LATEST_INFO = 'https://api.steampowered.com/IPublishedFileService/GetDetails/v1/?key={}&publishedfileids%5B0%5D={}&includemetadata=true&appid={}'

    def __init__(self):
        super().__init__()

    async def get_update(self, app_id: str, published_file_id: str, mods_dir: str, version: str = None, api_key: str = None) -> int:
        if not api_key:
            api_logger.error(f'{app_id}-{published_file_id}: The steam_api_key is null or empty and cannot get an '
                             f'update for the mod. Please provide a valid api key.')
        else:
            request = self._WORKSHOP_API_LATEST_INFO.format(api_key, published_file_id, app_id)
            api_logger.debug(f"Start API request: {request.replace(api_key, '<removed_key>')}")
            async with aiohttp.request('GET', request) as resp:
                if resp.status == 200:
                    j = await resp.json()

                    self.provider = 'workshop'
                    self.app = app_id
                    self.mods = published_file_id
                    self.mods_dir = mods_dir
                    self.version = j['response']['publishedfiledetails'][0]['time_updated']
                    self.title = j['response']['publishedfiledetails'][0]['title']
                    self.description = j['response']['publishedfiledetails'][0]['file_description']
                    self.full_mods_name = app_id + '-' + published_file_id

                    api_logger.debug(
                        f'The request from the "Thunderstore" API was successfull for the APP ID {app_id} and the published file ID {published_file_id}.')
                    return 0
                else:
                    api_logger.error(f'{app_id}-{published_file_id}: API Status {resp.status}, Error: {await resp.text()}')
                    return 1



import aiohttp
import logging
from vapordmods.api.base import BaseApi

api_logger = logging.getLogger(__name__)


class github(BaseApi):
    _GITHUB_API_RELEASE = 'https://api.github.com/repos/{}/{}/releases'

    def __init__(self):
        super().__init__()

    async def get_update(self, owner: str, repo: str, mods_dir: str, filename: str, version: str = None) -> int:

        payload = {'Accept': 'application/vnd.github.v3+json'}
        request = self._GITHUB_API_RELEASE.format(owner, repo)

        async with aiohttp.ClientSession() as session:
            async with session.get(request, params=payload) as resp:
                if resp.status == 200:
                    j = await resp.json()
                    release = []
                    if not version:
                        release = j[0]
                    else:
                        release = [x for x in j if x['tag_name'] == version]

                    if not len(release):
                        api_logger.error(f'{owner}-{repo}: No release found for version: {version}')
                        return 1

                    assets = [x for x in release['assets'] if x['name'].lower() == filename.lower()]
                    if not len(assets):
                        api_logger.error(f'{owner}-{repo}: No filename found for version: {filename}')
                        return 1

                    self.version = release['tag_name']
                    self.description = ''
                    self.provider = 'github'
                    self.app = owner
                    self.mods = repo
                    self.title = repo
                    self.mods_dir = mods_dir
                    self.full_mods_name = owner + '-' + repo
                    self.download_url = assets[0]['browser_download_url']

                    api_logger.debug(
                        f'The request from the "Thunderstore" API was successfull for the namespace {owner} and the mod {repo}')
                    return 0
                else:
                    api_logger.error(f'{owner}-{repo}: Status {resp.status}, Error: {await resp.text()}')
                    return 1



import aiohttp
import logging

from vapordmods.api.base import BaseApi

api_logger = logging.getLogger(__name__)


class nexusmods(BaseApi):
    _NEXUSMODS_API_URL_FILES = "https://api.nexusmods.com/v1/games/{}/mods/{}/files.json"
    _NEXUSMODS_API_URL_DOWNLOAD_LINK = "https://api.nexusmods.com/v1/games/{}/mods/{}/files/{}/download_link.json"

    def __init__(self):
        super().__init__()

    @staticmethod
    def __get_file_data(version, response):
        filedata = {}
        if not version:
            filedata = response['files'][len(response['files']) - 1]
        else:
            filedata = [i for i in response['files'] if i['version'] == version]
            if len(filedata) == 0:
                return None
            else:
                filedata = filedata[0]
        return filedata

    async def get_update(self, game_domain_name: str, mod_id: str, mods_dir: str, version: str = None, api_key: str = None) -> int:
        if not api_key:
            api_logger.error(f'{game_domain_name}-{mod_id}: The nmods_api_key is null or empty and cannot get an '
                             f'update for the mod. Please provide a valid api key.')
            return 1
        else:
            request = self._NEXUSMODS_API_URL_FILES.format(game_domain_name, mod_id)
            headers = {
                "accept": "application/json",
                "apikey": api_key
            }
            filedata = {}
            async with aiohttp.request('GET', request, headers=headers) as resp:
                if resp.status == 200:
                    j = await resp.json()

                    filedata = self.__get_file_data(version, j)
                    if filedata is None:
                        api_logger.error(
                            f"status': {resp.status}, 'data': The version '{version}' was not found in the game domain '{game_domain_name}' for the mod '{mod_id}'.")

                    req_dl = self._NEXUSMODS_API_URL_DOWNLOAD_LINK.format(game_domain_name, mod_id, str(filedata['file_id']))
                    async with aiohttp.request('GET', req_dl, headers=headers) as resp_dl:
                        if resp_dl.status == 200:
                            dl = await resp_dl.json()
                            self.app = game_domain_name
                            self.mods = mod_id
                            self.version = filedata['version']
                            self.download_url = dl[0]['URI']
                            return 0
            return 1



from steam.monkey import patch_minimal
patch_minimal()

import logging
import os
import asyncio
import aiofiles
import aiofiles.os
from builtins import staticmethod
from steam.client import SteamClient
from steam.client.cdn import CDNClient, CDNDepotFile
from steam.enums import EResult
from steam.client.builtins.web import webapi, make_requests_session
from steam.exceptions import ManifestError
from vapordmods.tools.utils import get_user_app_data

LOG = logging.getLogger(__name__)


class SteamManager:

    def __init__(self,
                 username: str,
                 password: str,
                 web_api_key: str = None,
                 steam_guard_code: str = None,
                 two_factor_code: str = None,
                 user_app_data_dir: str = None
                 ):
        if steam_guard_code and two_factor_code:
            LOG.error('steam_guard_code and two_factor_code are not None. You can only provide one of them.')
            raise SystemExit

        self.web_api_key = web_api_key
        self.client = client = SteamClient()
        self.username = username
        self.password = password
        self.steam_guard_code = steam_guard_code
        self.two_factor_code = two_factor_code

        self.user_app_data_dir = user_app_data_dir or os.path.join(get_user_app_data(), '.vapordmods')
        os.makedirs(self.user_app_data_dir, exist_ok=True)

        self.client.set_credential_location(self.user_app_data_dir)

        @client.on('error')
        def handle_error(result):
            LOG.info(f"Logon result: {result}")

        @client.on("connected")
        def handle_connected():
            LOG.info(f"Connected to {client.current_server_addr}")

        @client.on("logged_on")
        def handle_after_logon():
            LOG.info("Logged on as: %s", client.user.name)

        @client.on("disconnected")
        def handle_disconnect():
            LOG.info("Disconnected.")

    async def login(self):
        if not self.client.connected:
            if self.client.relogin_available:
                self.client.relogin()
                return 0

            elif self.username.lower() in ('anonymous', ''):
                result = self.client.anonymous_login()

            else:
                params = {
                    'username': self.username,
                    'password': self.password,
                    'auth_code': self.steam_guard_code,
                    'two_factor_code': self.two_factor_code
                }

                result = self.client.login(**params)

            if result != EResult.OK:
                return 1
        return 0

    async def search_workshop_item_manifest(self, published_file_id: int):
        params = {
            'key': self.web_api_key,
            'publishedfileids': [published_file_id],
            'includetags': 1,
            'includeadditionalpreviews': 1,
            'includechildren': 1,
            'includekvtags': 1,
            'includevotes': 1,
            'includeforsaledata': 1,
            'includemetadata': 1,
            'return_playtime_stats': 1,
            'strip_description_bbcode': 1,
        }

        try:
            result = webapi.get('IPublishedFileService',
                                'GetDetails',
                                params=params)['response']['publishedfiledetails'][0]
        except Exception as er:
            LOG.error(f'Query to the published file id {published_file_id} failed: {er}')
            return 1

        if result['result'] != EResult.OK:
            LOG.error(f"Query to the published file id {published_file_id} result : {EResult(result['result'])}")
            return 1

        return result

    @staticmethod
    async def _download_file_url(mods_dir, url, file):
        session = make_requests_session()
        stream = session.get(url, stream=True)

        ws_file = os.path.join(mods_dir, file)

        async with aiofiles.open(ws_file, 'wb') as f:
            for chunk in iter(lambda: stream.raw.read(8388608), b''):
                await f.write(chunk)

    async def _download_from_steampipe(self, mods_dir: str, pubfile: dict):
        async def dl_file(dfile: CDNDepotFile):
            file = dfile.read()
            path = os.path.join(mods_dir, pubfile['title'])
            await aiofiles.os.makedirs(path, exist_ok=True)

            filename = os.path.join(path, dfile.filename)
            async with aiofiles.open(filename, 'wb') as f:
                await f.write(file)

        if await self.login() == 0:
            key = pubfile['consumer_appid'], pubfile['consumer_appid'], pubfile['hcontent_file']
            cdn = CDNClient(self.client)
            try:
                manifest_code = cdn.get_manifest_request_code(*key)
                manifest = cdn.get_manifest(*key, manifest_request_code=manifest_code)
            except ManifestError as er:
                LOG.error(er)
                return 1

            to_dl = []
            for mfile in manifest:
                if not mfile.is_file:
                    continue
                to_dl.append(dl_file(mfile))
            await asyncio.gather(*to_dl)

    async def update_worksop_mod(self, mods_dir: str, published_file_id: int, pubfile: dict = None):
        if pubfile is None:
            pubfile = await self.search_workshop_item_manifest(published_file_id)
            if not pubfile:
                return 1

        if pubfile['result'] != EResult.OK:
            LOG.error('Error updating mods with invalid manifest.')
            return 1

        LOG.info(f"Updating the published file id {pubfile['publishedfileid']} for "
                 f"the app id {pubfile['consumer_appid']}.")

        if pubfile.get('file_url'):
            await self._download_file_url(mods_dir, pubfile['file_url'], pubfile['filename'])
        elif pubfile.get('hcontent_file'):
            await self._download_from_steampipe(mods_dir, pubfile)
        else:
            LOG.error(f"Cannot download the file for the  published file id {pubfile['publishedfileid']}")
            return 1

        return 0



schema = {
    'config': {
        'required': True,
        'type': 'dict',
        'schema': {
            'default_mods_dir': {
                'required': True,
                'nullable': True,
                'type': 'string',
            },
        },
    },
    'mods': {
        'required': True,
        'type': 'list',
        'schema': {
            'type': 'dict',
            'schema': {
                'provider': {
                    'required': True,
                    'type': 'string',
                    'allowed': ['thunderstore', 'nexusmods', 'workshop', 'github'],
                },
                'app': {
                    'required': True,
                },
                'mods': {
                    'required': True,
                },
                'version': {
                    'required': False,
                    'nullable': True,
                },
                'mods_dir': {
                    'required': False,
                    'nullable': True,
                    'type': 'string',
                },
                'filename': {
                    'required': False,
                    'nullable': False,
                    'type': 'string',
                },
            },
        },
    },
}



from abc import ABC, abstractmethod


class BaseApi(ABC):

    def __init__(self):
        self.provider = None
        self.app = None
        self.mods = None
        self.mods_dir = None
        self.version = None
        self.full_mods_name = None
        self.title = None
        self.description = None
        self.download_url = None

    @abstractmethod
    async def get_update(self, **kwargs) -> int:
        pass

    def return_data(self):
        return {'provider': self.provider,
                'app': self.app,
                'mods': self.mods,
                'mods_dir': self.mods_dir,
                'version': self.version,
                'full_mods_name': self.full_mods_name,
                'title': self.title,
                'description': self.description,
                'download_url': self.download_url
                }



from steam.monkey import patch_minimal

patch_minimal()

import shutil
import tempfile
import asyncio
import aiofiles
import aiofiles.os
import os
import aiohttp
import yaml
import logging
import zipfile
import pandas as pd
from cerberus import Validator
from pathlib import Path
from vapordmods.api import worhshop, thunderstore, nexusmods, github
from vapordmods.mods.schema import schema
from vapordmods.tools.steamcmd import SteamManager

logger = logging.getLogger(__name__)


class ModsManager:
    _CFG_FILENAME = 'vapordmods.yml'
    _MANIFESTS_FILENAME = 'vapordmods.manifests'
    _THUNDERSTORE_NAME = 'thunderstore'
    _NEXUSMODS_NAME = 'nexusmods'
    _WORKSHOP_NAME = 'workshop'
    _GITHUB_NAME = 'github'

    def __init__(self, install_dir: str, client: SteamManager = None):
        self.default_mods_dir = None
        self.install_dir = install_dir
        self.client = client
        self.manifests_filename = os.path.join(install_dir, self._MANIFESTS_FILENAME)
        self.cfg_filename = os.path.join(install_dir, self._CFG_FILENAME)
        self.cfg_data = {}
        self.mods_info = {}
        self.mods_status = {}

        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()

        if not os.path.exists(self.install_dir):
            raise NotADirectoryError(f"The installdir {self.install_dir} doesn't exist.")

        if not os.path.exists(self.cfg_filename) and not os.access(self.install_dir, os.W_OK):
            raise PermissionError(f"Cannot write to the folder {self.install_dir}.")

        if not os.path.exists(self.cfg_filename):
            loop.run_until_complete(self.write_cfg_filename())

    async def write_cfg_filename(self):
        template = b'config:\n  default_mods_dir: \n\nmods:\n  - provider: \n    app: \n    mods: \n    version: \n' \
                   b'    \n  - provider:             \n    app:       \n    mods:              \n    version: \n'
        async with aiofiles.open(self.cfg_filename, 'wb') as cfg_file:
            await cfg_file.write(template)

    @staticmethod
    async def __load_yaml(filename):
        if await aiofiles.os.path.exists(filename):
            async with aiofiles.open(filename, 'r') as file:
                return yaml.safe_load(await file.read())
        else:
            raise FileExistsError(filename)

    async def load_cfg_data(self):
        cfg_data = await self.__load_yaml(self.cfg_filename)
        mods_validator = Validator(schema)

        try:
            if not mods_validator.validate(cfg_data):
                logger.error(mods_validator.errors)
                return False
        except Exception as er:
            logger.error(er)
            return False

        self.cfg_data = cfg_data
        self.default_mods_dir = cfg_data['config']['default_mods_dir'] or self.install_dir
        return True

    async def load_mods_info(self):
        if os.path.exists(self.manifests_filename):
            self.mods_info = await self.__load_yaml(self.manifests_filename)

    def get_mods_info(self):
        return self.mods_info

    def get_mods_status(self):
        return self.mods_status

    async def refresh_mods_info(self, nmods_api_key: str = None, steam_api_key: str = None):
        try:
            suffixes = '_current'
            await self.load_cfg_data()
            self.mods_info = {}
            await self.load_mods_info()

            df_cfg = pd.DataFrame.from_dict(self.cfg_data['mods'])
            for col in ['version', 'mods_dir']:
                if col not in df_cfg.columns:
                    df_cfg[col] = ''

            df_cfg = df_cfg.fillna('')
            df_cfg.loc[((df_cfg['mods_dir'] == '') | (df_cfg['mods_dir'] is None)), 'mods_dir'] = self.default_mods_dir

            # Requests mods update
            mods_update = []
            apicall = None
            list_api_key = {self._THUNDERSTORE_NAME: None, self._NEXUSMODS_NAME: nmods_api_key,
                            self._WORKSHOP_NAME: steam_api_key, self._GITHUB_NAME: None}
            for idx, row in df_cfg.iterrows():
                apicall = getattr(globals()[row['provider']], row['provider'])()
                if row['provider'] in (self._THUNDERSTORE_NAME, self._NEXUSMODS_NAME, self._WORKSHOP_NAME):
                    params = {
                        'namespace': row['app'],
                        'name': row['mods'],
                        'mods_dir': row['mods_dir'],
                        'version': row['version'],
                        'api_key': list_api_key[row['provider']]
                    }
                else:
                    params = {
                        'owner': row['app'],
                        'repo': row['mods'],
                        'mods_dir': row['mods_dir'],
                        'version': row['version'],
                        'filename': row['filename']
                    }

                if await apicall.get_update(**params) == 0:
                    mods_update.append(apicall.return_data())

            df_update = pd.DataFrame(mods_update)
            df_update['need_update'] = False

            if len(self.mods_info):
                df_current = pd.DataFrame(self.mods_info)
                df_status = df_update.merge(df_current, on=['provider', 'app', 'mods'], suffixes=(None, suffixes))
                df_status['need_update'] = df_status['version'] != df_status[f'version{suffixes}']
            else:
                df_status = df_update
                df_status['need_update'] = True

            self.mods_status = df_status.to_dict('records')
            return 1
        except Exception as er:
            logger.error(f"Error during update for mods: {er}")
            return 0

    @staticmethod
    async def __extract_mods(filename, destination, row):
        with zipfile.ZipFile(filename, 'r') as file:
            temp_dir = os.path.join(tempfile.gettempdir(), 'vapord_temp')
            try:
                if row['provider'].lower() == 'thunderstore' and 'bepinexpack' in row['full_mods_name'].lower():
                    benpinex = 'BepInEx'
                    file_list = [x for x in file.namelist() if benpinex in x]
                    benpinex_dir = file_list[0][0:file_list[0].index(os.sep)]
                    temp_dir_bepinex = os.path.join(temp_dir, benpinex_dir)
                    file.extractall(temp_dir, file_list)
                    cp_dir = temp_dir_bepinex
                else:
                    file.extractall(temp_dir)
                    cp_dir = temp_dir

                shutil.copytree(cp_dir, row['mods_dir'], dirs_exist_ok=True)

                return 1
            except Exception as er:
                logger.error(er)
                return 0
            finally:
                if await aiofiles.os.path.exists(temp_dir):
                    shutil.rmtree(temp_dir)

    async def __make_request(self, session, row):
        try:
            resp = await session.request(method="GET", url=row['download_url'])

            if resp.status == 200:
                await aiofiles.os.makedirs(row['mods_dir'], exist_ok=True)

                filename = os.path.join(row['mods_dir'], resp.url.name)
                destination = os.path.join(row['mods_dir'], row['full_mods_name'])
                async with aiofiles.open(filename, 'wb') as f:
                    await f.write(await resp.read())

                if await self.__extract_mods(filename, destination, row):
                    await aiofiles.os.remove(filename)
            else:
                logger.error(f"Error with the request: {resp.status} {resp.text()}")

        except Exception as er:
            logger.error(er)

    async def update_mods(self):
        if not len(self.mods_status):
            logger.error(f"No mods information. Please execute the method 'refresh_mods_info'.")
            return 0

        try:
            list_to_update = [x for x in self.mods_status if x['need_update'] and x['provider'] in
                              [self._THUNDERSTORE_NAME, self._NEXUSMODS_NAME, self._GITHUB_NAME]]

            if len(list_to_update):
                async with aiohttp.ClientSession() as session:
                    tasks = [self.__make_request(session, i) for i in list_to_update]

                    await asyncio.gather(*tasks)

            list_to_update_workshop = [x for x in self.mods_status if
                                       x['need_update'] == 'True' and x['provider'] == self._WORKSHOP_NAME]
            if len(list_to_update_workshop):
                for i in list_to_update_workshop:
                    result = await self.client.update_worksop_mod(i['app'], i['mods'])
                    if result == 1 and result(1):
                        Path(result(1)).symlink_to(os.path.join(i['mods_dir'], i['title']))
                    elif result == 1:
                        logger.error(
                            f"The symlink for the APP_ID {i['app']} and published_file_id {i['mods']} was note created.")
                    else:
                        logger.error(result(1))

            with open(self.manifests_filename, 'w') as manifest:
                yaml.safe_dump(self.mods_status, manifest)

            return 1
        except Exception as er:
            logger.error(er)
            return 0
